import datetime
import functools
import getpass
import importlib.metadata
import socket
import sys

from elliptio.interfaces import AutomaticMetadata, TrackerInterface


class Tracker(TrackerInterface):
    def get_automatic_metadata(self) -> AutomaticMetadata:
        return AutomaticMetadata(
            creation_time=datetime.datetime.now(tz=datetime.timezone.utc),
            argv=" ".join(sys.argv),  # NOTE: before sys.orig_argv using 3.11 ?!
            username=_get_username(),
            hostname=_get_hostname(),
            python_packages=_get_python_packages(),
        )


@functools.lru_cache
def _get_hostname():
    return socket.gethostname()


@functools.lru_cache
def _get_username():
    return getpass.getuser()


def _get_python_packages():
    """Gets list of installed python packages.

    Some alternatives:
    - `pip freeze`: pip is not installed in rye or pdm venvs
    - `pkg_resources`: deprecated in favor of importlib.resources or importlib.metadata
    """
    return {d.metadata["name"]: d.version for d in importlib.metadata.distributions()}



from __future__ import annotations

import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import datetime

    import pandas as pd

from dataclasses import asdict, dataclass

import yaml


class ID(str):
    __slots__ = ()


@dataclass
class AutomaticMetadata:
    creation_time: datetime.datetime
    argv: str
    username: str
    hostname: str
    python_packages: dict[str, str]  # maybe also as JSON string?!

    # TODO
    # git hash, git diff, ...
    # machine hostname (IP-address?)
    # env-vars (caution: passwords!)


@dataclass
class ManualMetadata:
    ticket: str = ""
    project: str = ""
    datatype: str = ""
    dataset: str = ""
    description: str = ""
    config: str = ""  # e.g.  JSON or YAML string


@dataclass
class FileInfo:
    file_id: ID
    run_id: ID
    relpath: str
    remote_url: str
    file_hash: str
    byte_size: int

    automatic_metadata: AutomaticMetadata
    manual_metadata: ManualMetadata

    deduplicated_by: ID | None
    based_on: list[ID]

    def to_yaml(self):
        dct = asdict(self)
        _deep_map(dct, _convert_ids_to_str)
        return yaml.safe_dump(dct)


def _convert_ids_to_str(x):
    if isinstance(x, ID):
        return str(x)
    if isinstance(x, list) and len(x) > 0 and isinstance(x[0], ID):
        return [str(e) for e in x]
    return x


def _deep_map(nested_dict, func):
    for key, value in nested_dict.items():
        if isinstance(value, dict):
            _deep_map(value, func)
        else:
            nested_dict[key] = func(value)


@dataclass
class FileSystemInterface(abc.ABC):
    @abc.abstractmethod
    def define_remote_url(
        self,
        amd: AutomaticMetadata,
        file_id: ID,
        relpath: str,
    ) -> str:
        pass

    @abc.abstractmethod
    def create_parent_dir(self, remote_url: str) -> None:
        pass

    @abc.abstractmethod
    def upload(self, local_path: str, remote_url: str) -> None:
        pass

    @abc.abstractmethod
    def download(self, remote_url: str, local_path: str) -> None:
        pass

    @abc.abstractmethod
    def write_text(self, remote_url: str, text: str) -> None:
        pass

    @abc.abstractmethod
    def read_text(self, remote_url: str) -> str:
        pass

    @abc.abstractmethod
    def get_hash(self, remote_url: str) -> str:
        pass

    @abc.abstractmethod
    def get_byte_size(self, remote_url: str) -> int:
        pass


@dataclass
class File(FileInfo):
    fs: FileSystemInterface

    def upload(self, local_path: str) -> None:
        self.fs.upload(local_path, self.remote_url)

    def download(self, local_path: str) -> None:
        self.fs.download(self.remote_url, local_path)

    def write_text(self, text: str) -> None:
        self.fs.write_text(self.remote_url, text)

    def read_text(self) -> str:
        return self.fs.read_text(self.remote_url)

    def update_hash_and_byte_size(self):
        self.file_hash = self.fs.get_hash(self.remote_url)
        self.byte_size = self.fs.get_byte_size(self.remote_url)


@dataclass
class DataBaseInterface(abc.ABC):
    @abc.abstractmethod
    def save(self, info: FileInfo) -> None:
        pass

    @abc.abstractmethod
    def load(self, file_id: ID) -> FileInfo:
        pass

    @abc.abstractmethod
    def query(self, filter_dict: dict | None, **kwargs) -> pd.DataFrame:
        # The dataframe has columns corresponding to FileInfo
        # TODO: Can we create a new type for this, e.g. DataFrame[FileInfo]?
        pass


class IdCreatorInterface(abc.ABC):
    @abc.abstractmethod
    def create_unique_id(self, prefix: str = "") -> ID:
        pass


@dataclass
class TrackerInterface(abc.ABC):
    @abc.abstractmethod
    def get_automatic_metadata(self) -> AutomaticMetadata:
        pass



import uuid

from elliptio.interfaces import ID, IdCreatorInterface


class IdCreator(IdCreatorInterface):
    def create_unique_id(self, prefix: str = "") -> ID:
        id_ = uuid.uuid4()
        if not prefix:
            return ID(id_)
        return ID(f"{prefix}_{id_}")



from __future__ import annotations

import contextlib
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Iterator

if TYPE_CHECKING:
    import pandas as pd

from .interfaces import (
    ID,
    AutomaticMetadata,
    DataBaseInterface,
    File,
    FileSystemInterface,
    IdCreatorInterface,
    ManualMetadata,
    TrackerInterface,
)


@dataclass
class Handler:
    fs: FileSystemInterface
    db: DataBaseInterface
    tracker: TrackerInterface
    id_creator: IdCreatorInterface
    based_on: list[ID] = field(default_factory=list)
    mmd: ManualMetadata | None = None

    def __post_init__(self):
        self.run_id = self.id_creator.create_unique_id(prefix="run")

    @contextlib.contextmanager
    def create(
        self,
        relpath: str,
        *,
        manual_metadata: ManualMetadata | None = None,
        automatic_metadata: AutomaticMetadata | None = None,
        save_metadata_to_filesystem: bool = True,
    ) -> Iterator[File]:
        amd = automatic_metadata or self.tracker.get_automatic_metadata()
        mmd = manual_metadata or self.mmd or ManualMetadata()
        file_id = self.id_creator.create_unique_id("file")
        remote_url = self.fs.define_remote_url(amd, file_id, relpath)
        self.fs.create_parent_dir(remote_url)
        file = File(
            file_id=file_id,
            run_id=self.run_id,
            relpath=relpath,
            remote_url=remote_url,
            file_hash="",
            byte_size=0,
            automatic_metadata=amd,
            manual_metadata=mmd,
            deduplicated_by=None,
            based_on=self.based_on.copy(),
            fs=self.fs,
        )

        yield file
        file.update_hash_and_byte_size()
        self.db.save(file)
        if save_metadata_to_filesystem:
            self.fs.write_text(
                remote_url=self.fs.define_remote_url(amd, file_id, ".metadata.yaml"),
                text=file.to_yaml(),
            )

    def load(self, file_id: ID) -> File:
        file_info = self.db.load(file_id)
        while file_info.deduplicated_by:
            file_info = self.db.load(file_info.deduplicated_by)
        self.based_on.append(file_info.file_id)
        return File(**asdict(file_info), fs=self.fs)

    def query(self, dct: dict | None = None, **kwargs) -> pd.DataFrame:
        # The type of query depends on the database
        dct = {} if dct is None else dct
        return self.db.query(dct, **kwargs)

    def deduplicate(self):
        # TODO: I think this shouldn't be too difficult, but done for today.
        raise NotImplementedError



from __future__ import annotations

import contextlib
import json
from dataclasses import dataclass

# NOTE: This cannot go into the type checking block, otherwise SQL runtime error
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Iterator, Optional

if TYPE_CHECKING:
    from pathlib import Path

import pandas as pd
import sqlalchemy
from sqlalchemy.exc import NoResultFound  # type: ignore[attr-defined]
from sqlalchemy.orm import (  # type: ignore[attr-defined]
    DeclarativeBase,
    Mapped,
    MappedAsDataclass,
    Session,
    mapped_column,
)

from elliptio.interfaces import (
    ID,
    AutomaticMetadata,
    DataBaseInterface,
    FileInfo,
    ManualMetadata,
)


class NoDatabaseEntryError(ValueError):
    pass


@dataclass
class SqlDatabase(DataBaseInterface):
    def __init__(self, db_path: Path) -> None:
        self.engine = sqlalchemy.create_engine(f"sqlite:///{db_path}")
        _Base.metadata.create_all(self.engine)

    def save(self, info: FileInfo) -> None:
        row = _FileRow.from_fileinfo(info)
        with self._get_session() as s:
            s.add(row)
            s.commit()

    def load(self, file_id: ID) -> FileInfo:
        with self._get_session() as s:
            try:
                row: _FileRow = s.query(_FileRow).get(file_id)  # type: ignore[assignment]
            except NoResultFound:
                raise NoDatabaseEntryError from None
            else:
                return row.to_fileinfo()

    def query(self, filter_dict: dict | None, **kwargs) -> pd.DataFrame:
        del kwargs  # we don't need it here, but other adapters might!
        dct = filter_dict or {}
        with self._get_session() as s:
            query = s.query(_FileRow).filter_by(**dct)
            return pd.read_sql(query.statement, query.session.bind)  # type: ignore [arg-type]

    @contextlib.contextmanager
    def _get_session(self) -> Iterator[Session]:
        with Session(self.engine) as s:  # type: ignore[attr-defined]
            yield s


class _Base(DeclarativeBase):
    pass


# NOTE: Can simplify this?! See https://docs.sqlalchemy.org/en/20/orm/dataclasses.html
class _FileRow(_Base, MappedAsDataclass):
    __tablename__ = "files"

    file_id: Mapped[str] = mapped_column(primary_key=True)
    run_id: Mapped[str]
    relpath: Mapped[str]
    remote_url: Mapped[str]
    file_hash: Mapped[str]
    byte_size: Mapped[int]
    deduplicated_by: Mapped[Optional[str]]  # noqa: UP007
    based_on: Mapped[str]

    # automatic metadata
    creation_time: Mapped[datetime]
    argv: Mapped[str]
    username: Mapped[str]
    hostname: Mapped[str]
    python_packages: Mapped[str]

    # manual_metadata
    ticket: Mapped[str]
    project: Mapped[str]
    datatype: Mapped[str]
    dataset: Mapped[str]
    description: Mapped[str]
    config: Mapped[str]

    @classmethod
    def from_fileinfo(cls, info: FileInfo) -> _FileRow:
        amd = info.automatic_metadata
        mmd = info.manual_metadata
        return _FileRow(
            file_id=info.file_id,
            run_id=info.run_id,
            relpath=info.relpath,
            remote_url=info.remote_url,
            file_hash=info.file_hash,
            byte_size=info.byte_size,
            deduplicated_by=info.deduplicated_by,
            based_on=json.dumps(info.based_on),
            # automatic metadata
            creation_time=amd.creation_time.astimezone(None),
            argv=amd.argv,
            username=amd.username,
            hostname=amd.hostname,
            python_packages=json.dumps(amd.python_packages),
            # manual metadata
            ticket=mmd.ticket,
            project=mmd.project,
            datatype=mmd.datatype,
            dataset=mmd.dataset,
            description=mmd.description,
            config=mmd.config,
        )

    def to_fileinfo(self) -> FileInfo:
        amd = AutomaticMetadata(
            creation_time=self.creation_time.astimezone(tz=timezone.utc),
            argv=self.argv,
            username=self.username,
            hostname=self.hostname,
            python_packages=json.loads(self.python_packages),
        )
        mmd = ManualMetadata(
            ticket=self.ticket,
            project=self.project,
            datatype=self.datatype,
            dataset=self.dataset,
            description=self.description,
            config=self.config,
        )
        return FileInfo(
            file_id=ID(self.file_id),
            run_id=ID(self.run_id),
            relpath=self.relpath,
            remote_url=self.remote_url,
            file_hash=self.file_hash,
            byte_size=self.byte_size,
            automatic_metadata=amd,
            manual_metadata=mmd,
            deduplicated_by=_optional_str_to_id(self.deduplicated_by),
            based_on=json.loads(self.based_on),
        )


def _optional_str_to_id(s: str | None) -> ID | None:
    if s is None:
        return None
    return ID(s)



import hashlib

import fsspec

from elliptio.interfaces import ID, AutomaticMetadata, FileSystemInterface


class FsspecFilesystem(FileSystemInterface):
    def __init__(self, prefix: str, protocol="file", **storage_options) -> None:
        self.fs: fsspec.AbstractFileSystem = fsspec.filesystem(
            protocol,
            **storage_options,
        )
        self.prefix = prefix

    def define_remote_url(
        self,
        amd: AutomaticMetadata,
        file_id: ID,
        relpath: str,
    ) -> str:
        return self.fs.sep.join(
            [
                self.prefix,
                str(amd.creation_time.year),
                str(amd.creation_time.month),
                str(amd.creation_time.day),
                str(amd.username),
                file_id,
                relpath,
            ],
        )

    def create_parent_dir(self, remote_url: str) -> None:
        self.fs.mkdirs(self.fs._parent(remote_url), exist_ok=True)  # noqa: SLF001

    def upload(self, local_path: str, remote_url: str) -> None:
        self.create_parent_dir(remote_url)
        self.fs.put(lpath=local_path, rpath=remote_url)

    def download(self, remote_url: str, local_path: str) -> None:
        self.fs.get(rpath=remote_url, lpath=local_path)

    def write_text(self, remote_url: str, text: str) -> None:
        self.create_parent_dir(remote_url)
        with self.fs.open(remote_url, "w") as f:
            f.write(text)

    def read_text(self, remote_url: str) -> str:
        with self.fs.open(remote_url, "r") as r:
            return r.read()

    def get_hash(self, remote_url: str) -> str:
        hasher = hashlib.sha256()
        with self.fs.open(remote_url, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hasher.update(chunk)
        return hasher.hexdigest()

    def get_byte_size(self, remote_url: str) -> int:
        with self.fs.open(remote_url, "rb") as remote_file:
            remote_file.seek(0, 2)  # Move to the end of the file
            return remote_file.tell()  # Current position (in bytes)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from elliptio.adapters.db import SqlDatabase
from elliptio.adapters.fs import FsspecFilesystem
from elliptio.adapters.id_creator import IdCreator
from elliptio.adapters.tracker import Tracker
from elliptio.handler import Handler

if TYPE_CHECKING:
    from elliptio.interfaces import (
        DataBaseInterface,
        FileSystemInterface,
        IdCreatorInterface,
        ManualMetadata,
        TrackerInterface,
    )


class DirectoryDoesNotExistError(ValueError):
    def __init__(self, dirpath: Path | str) -> None:
        super().__init__(f"{dirpath=} does not exist.")


def get_handler(  # noqa: PLR0913
    *,
    dirpath: Path | str | None = None,
    fs: FileSystemInterface | None = None,
    db: DataBaseInterface | None = None,
    tracker: TrackerInterface | None = None,
    id_creator: IdCreatorInterface | None = None,
    manual_metadata: ManualMetadata | None = None,
) -> Handler:
    if dirpath is not None and not Path(dirpath).exists():
        raise DirectoryDoesNotExistError(dirpath)

    return Handler(
        fs=fs or FsspecFilesystem(str(dirpath)),
        db=db or SqlDatabase(Path(dirpath) / "db.sqlite"),  # type: ignore [arg-type]
        tracker=tracker or Tracker(),
        id_creator=id_creator or IdCreator(),
        mmd=manual_metadata,
    )
